""" Local resolver API
"""
from core_linked_records_app.settings import PID_XPATH
from core_main_app.commons.exceptions import ApiError, DoesNotExist
from core_main_app.components.data import api as data_api


def get_data_by_pid(pid, user):
    """ Return data object with the given pid.

        Parameters:
            pid:
            user:

        Returns: data object
    """
    json_pid_path = "dict_content.%s" % PID_XPATH
    query_result = data_api.execute_query({json_pid_path: pid}, user)
    query_result_length = len(query_result)

    if query_result_length == 0:
        raise DoesNotExist("No result found")
    elif query_result_length != 1:
        raise ApiError("PID must be unique")
    else:
        return query_result[0]



""" Data Xml renderer for django REST API
"""
from django.http import HttpResponse
from rest_framework import renderers
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_500_INTERNAL_SERVER_ERROR


class DataXmlRenderer(renderers.BaseRenderer):
    media_type = 'application/xml'
    format = 'xml'
    charset = 'utf-8'

    def render(self, data, media_type=None, renderer_context=None):
        """ Render the data object by just returning the xml_content field

        Args:
            data:
            media_type:
            renderer_context:

        Returns: xml string
        """

        # check errors
        if 'status' in data and data['status'] == 'error':
            if data['code'] == HTTP_404_NOT_FOUND:
                return HttpResponse(status=HTTP_404_NOT_FOUND)
            else:
                return HttpResponse(status=HTTP_500_INTERNAL_SERVER_ERROR)
        else:
            return data['xml_content']



""" System API for core_linked_records
"""
from core_linked_records_app.settings import PID_XPATH
from core_main_app.components.data.models import Data


def is_pid_defined_for_document(pid, document_id):
    """ Determine if a given PID match the document ID provided.

    Params:
        pid:
        document_id:

    Returns:
    """
    json_pid_path = "dict_content.%s" % PID_XPATH
    query_result = Data.execute_query({json_pid_path: pid}, order_by_field=[])

    return len(query_result) == 1 and query_result[0].pk == document_id


def is_pid_defined(pid):
    """ Determine if a given PID already exists.

    Params:
        pid:

    Returns:
    """
    json_pid_path = "dict_content.%s" % PID_XPATH
    query_result = Data.execute_query({json_pid_path: pid}, order_by_field=[])

    return len(query_result) == 1



""" Data HTML renderer for django REST API
"""
import logging
from django.http import HttpResponse
from rest_framework import renderers
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_500_INTERNAL_SERVER_ERROR

from core_main_app.components.data import api as data_api
from core_main_app.utils.rendering import render

LOGGER = logging.getLogger(__name__)


class DataHtmlUserRenderer(renderers.BaseRenderer):
    media_type = 'text/html'
    format = 'html'
    charset = 'utf-8'

    @staticmethod
    def build_page(data, request):
        context = {
            'data': data_api.get_by_id(data["id"], request.user)
        }

        assets = {
            "js": [
                {
                    "path": 'core_main_app/common/js/XMLTree.js',
                    "is_raw": False
                },
                {
                    "path": 'core_main_app/user/js/data/detail.js',
                    "is_raw": False
                },
            ],
            "css": ["core_main_app/common/css/XMLTree.css"],
        }

        # check errors
        if 'status' in data and data['status'] == 'error':
            if data['code'] == HTTP_404_NOT_FOUND:
                return HttpResponse(status=HTTP_404_NOT_FOUND)
            else:
                return HttpResponse(status=HTTP_500_INTERNAL_SERVER_ERROR)
        else:
            return render(
                request, 'core_main_app/user/data/detail.html', context=context,
                assets=assets
            )

    def render(self, data, media_type=None, renderer_context=None):
        """ Render the data object by returning the user template

        Args:
            data:
            media_type:
            renderer_context:

        Returns: html page
        """
        try:
            return self.build_page(data, renderer_context["request"])
        except Exception as e:
            LOGGER.error("Error while building data page: %s" % str(e))

            if "kwargs" in renderer_context and "record" in renderer_context["kwargs"]:
                error_msg = "Document %s does not exist." % \
                            renderer_context["kwargs"]["record"]
            else:
                error_msg = "Invalid request provided."

            return render(
                renderer_context["request"], 'core_main_app/common/commons/error.html',
                context={
                    "error": error_msg
                }
            )



""" Url router for the core linked records app
"""
from django.conf.urls import url

from core_linked_records_app.rest.providers import views as providers_views

urlpatterns = [
    url(r'^(?P<provider>[^/]+)/(?P<record>.*)$',
        providers_views.ProviderRecord.as_view(),
        name='core_linked_records_app_rest_provider_record_view')
]



""" Apps file for setting linked records when app is ready
"""
import sys

from django.apps import AppConfig

from core_linked_records_app.components.data import watch as data_watch


class LinkedRecordsAppConfig(AppConfig):
    """ Core application settings
    """
    name = 'core_linked_records_app'

    def ready(self):
        """ Run when the app is ready

        Returns:

        """
        if 'migrate' not in sys.argv:
            data_watch.init()



""" Handle system abstract class
"""
from abc import ABC, abstractmethod


class AbstractIdProvider(ABC):
    def __init__(self, base_url, cdcs_base_url, username, password):
        self.provider_url = base_url
        self.local_url = cdcs_base_url
        self.auth_token = self.encode_token(username, password)

    @abstractmethod
    def encode_token(self, username, password):
        raise NotImplementedError()

    @abstractmethod
    def get(self, record):
        raise NotImplementedError()

    @abstractmethod
    def create(self, prefix, record=None):
        raise NotImplementedError()

    @abstractmethod
    def update(self, record):
        raise NotImplementedError()

    @abstractmethod
    def delete(self, record):
        raise NotImplementedError()



""" Linked records REST views
"""
import json
import logging
from importlib import import_module

from rest_framework.parsers import JSONParser
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework.status import HTTP_404_NOT_FOUND, HTTP_500_INTERNAL_SERVER_ERROR, HTTP_200_OK
from rest_framework.views import APIView

from core_linked_records_app.components.data.api import get_data_by_pid
from core_linked_records_app.rest.data.renderers.data_html_user_renderer import DataHtmlUserRenderer
from core_linked_records_app.rest.data.renderers.data_xml_renderer import DataXmlRenderer
from core_linked_records_app.settings import ID_PROVIDER_SYSTEMS, ID_PROVIDER_PREFIX_DEFAULT, \
    ID_PROVIDER_PREFIXES
from core_main_app.commons.exceptions import DoesNotExist
from core_main_app.rest.data.serializers import DataSerializer

LOGGER = logging.getLogger(__name__)


class ProviderRecord(APIView):
    parser_classes = (JSONParser,)
    renderer_classes = (DataHtmlUserRenderer, JSONRenderer, DataXmlRenderer)

    def __init__(self):
        self.id_provider_instances = {
            id_provider: None for id_provider in ID_PROVIDER_SYSTEMS.keys()
        }
        super().__init__()

    def _get_provider_instance(self, system):
        if self.id_provider_instances[system] is None:
            # Retrieve class name and module path
            id_provider_classpath = ID_PROVIDER_SYSTEMS[system]["class"].split(".")
            id_provider_classname = id_provider_classpath[-1]
            id_provider_modpath = ".".join(id_provider_classpath[:-1])

            # Import module and class
            id_provider_module = import_module(id_provider_modpath)
            id_provider_class = getattr(
                id_provider_module, id_provider_classname
            )

            # Initialize handel system instance
            self.id_provider_instances[system] = id_provider_class(
                *ID_PROVIDER_SYSTEMS[system]["args"]
            )

        return self.id_provider_instances[system]

    def post(self, request, provider, record):
        """ Create a handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        # Parse record entry to split between prefix and record ID.
        prefix_record_list = record.split("/")

        if len(prefix_record_list) == 1:  # No record provided
            prefix = prefix_record_list[0]
            record = None
        elif prefix_record_list[-1] == "":  # Only prefix provided
            prefix = "/".join(prefix_record_list[:-1])
            record = None
        else:  # Prefix and record provided
            prefix = "/".join(prefix_record_list[:-1])
            record = prefix_record_list[-1]

        # Assign default prefix if the prefix is undefined or not in the list of
        # authorized ones.
        if prefix == "" or prefix not in ID_PROVIDER_PREFIXES:
            prefix = ID_PROVIDER_PREFIX_DEFAULT

        id_provider = self._get_provider_instance(provider)
        provider_response = id_provider.create(prefix, record)

        provider_content = json.loads(provider_response.content)
        return Response(
            provider_content,
            status=provider_response.status_code
        )

    def put(self, request, provider, record):
        """ Update the value of a given handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        id_provider = self._get_provider_instance(provider)
        provider_response = id_provider.update(record)

        provider_content = json.loads(provider_response.content)

        return Response(
            provider_content,
            status=provider_response.status_code
        )

    def get(self, request, provider, record):
        """ Retrieve the local data of a given handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        id_provider = self._get_provider_instance(provider)
        provider_response = id_provider.get(record)

        try:
            query_result = get_data_by_pid(
                json.loads(provider_response.content)["url"], request.user
            )
            return Response(
                DataSerializer(query_result).data,
                status=HTTP_200_OK
            )
        except DoesNotExist:
            content = {
                "status": "error",
                "code": HTTP_404_NOT_FOUND,
                "message": "No data with specified handle found"
            }
            return Response(content, status=HTTP_404_NOT_FOUND)
        except Exception as ex:
            content = {
                "status": "error",
                "code": HTTP_500_INTERNAL_SERVER_ERROR,
                "message": str(ex)
            }
            return Response(
                content, status=HTTP_500_INTERNAL_SERVER_ERROR
            )

    def delete(self, request, provider, record):
        """ Delete a handle record

        Args:
            request:
            provider:
            record:

        Returns:
        """
        id_provider = self._get_provider_instance(provider)
        provider_response = id_provider.delete(record)

        provider_content = json.loads(provider_response.content)

        return Response(
            provider_content,
            status=provider_response.status_code
        )



""" Core linked records app settings
"""
from django.conf import settings

if not settings.configured:
    settings.configure()

SERVER_URI = getattr(settings, "SERVER_URI", "http://localhost")

PID_XPATH = getattr(settings, "PID_XPATH", "Resource.@localid")

PID_FORMAT = getattr(settings, "PID_FORMAT", r"[a-zA-Z0-9_\-]+")

ID_PROVIDER_SYSTEMS = getattr(settings, "ID_PROVIDER_SYSTEMS", {
    "local": {
        "class":
            "core_linked_records_app.utils.providers.local.LocalIdProvider",
        "args": [
            SERVER_URI
        ]
    }
})

ID_PROVIDER_PREFIXES = getattr(settings, "ID_PROVIDER_PREFIXES", ["cdcs"])

ID_PROVIDER_PREFIX_DEFAULT = getattr(settings, "ID_PROVIDER_PREFIX_DEFAULT",
                                     ID_PROVIDER_PREFIXES[0])



""" Signals to trigger before Data save
"""
import logging
import re

from django.urls import reverse

from core_linked_records_app.settings import ID_PROVIDER_SYSTEMS, PID_XPATH, SERVER_URI, \
    ID_PROVIDER_PREFIXES, PID_FORMAT
from core_linked_records_app.system import api as system_api
from core_linked_records_app.utils.xml import get_xpath_from_dot_notation, \
    get_xpath_with_target_namespace, get_value_at_xpath, set_value_at_xpath
from core_main_app.commons import exceptions
from core_main_app.components.data.models import Data
from core_main_app.utils.requests_utils.requests_utils import send_post_request
from signals_utils.signals.mongo import signals, connector
from xml_utils.xsd_tree.xsd_tree import XSDTree

LOGGER = logging.getLogger(__name__)


def init():
    """ Connect to Data object events.
    """
    connector.connect(set_data_pid, signals.pre_save, sender=Data)


def set_data_pid(sender, document, **kwargs):
    """ Set the PID in the XML field specified in the settings. If the PID
    already exists and is valid, it is not reset.

    Params:
        sender:
        document:
        kwargs:

    Returns:
    """
    # FIXME remove hard-coded variables
    default_system = list(ID_PROVIDER_SYSTEMS.keys())[0]

    pid_xpath = get_xpath_from_dot_notation(PID_XPATH)

    # Retrieve namespaces
    xml_tree = XSDTree.build_tree(document.xml_content)
    pid_xpath, namespaces = get_xpath_with_target_namespace(
        pid_xpath, document.template.content
    )

    try:  # Get the PID from the `pid_xpath` value
        document_pid = get_value_at_xpath(xml_tree, pid_xpath, namespaces)
        if type(document_pid) == str and document_pid.endswith("/"):  # Cleanup PID if it ends with a '/'
            document_pid = document_pid[:-1]
    except AssertionError:  # If PID XPath not found in document, do not go further
        return

    # Generate the default URL for the default system and remove the final '/'
    pid_generation_url = ("%s%s" % (
        SERVER_URI,
        reverse(
            "core_linked_records_app_rest_provider_record_view",
            kwargs={
                "provider": default_system,
                "record": ""
            }
        )
    ))[:-1]

    # Generate new PID if none has been specified or need prefix and record generated
    if document_pid is None or document_pid == "":
        document_pid = "%s/" % pid_generation_url
    elif re.match(r"^%s(/%s)?$" % (pid_generation_url, PID_FORMAT), document_pid) is not None:
        # PID specified with only prefix (record to be generated)
        document_pid = "%s/" % document_pid
    elif re.match(r"^%s/%s/%s$" % (pid_generation_url, PID_FORMAT, PID_FORMAT),
                  document_pid) is not None:  # PID has been specified (prefix + record)
        # Check that PID is not assigned to another document
        # * Check 1: document is already registered in database
        if document.pk is not None \
                and not system_api.is_pid_defined_for_document(document_pid, document.pk) \
                and system_api.is_pid_defined(document_pid):
            raise exceptions.ModelError("PID already defined for another document")

        # * Check 2: registering a new document in the database
        if document.pk is None and system_api.is_pid_defined(document_pid):
            raise exceptions.ModelError("PID already defined for another document")
    else:
        # PID specified but not matching any possible URLs for the generation
        raise exceptions.ModelError("Invalid PID provided")

    # Register the PID and return the URL provided
    document_pid_response = send_post_request("%s?format=json" % document_pid)
    document_pid = document_pid_response.json()["url"]

    # Set the document PID into XML data and update `xml_content`
    set_value_at_xpath(xml_tree, pid_xpath, document_pid, namespaces)
    document.xml_content = XSDTree.tostring(xml_tree)

    # Update the whole document with the updated XML content
    document.convert_to_file()
    document.convert_to_dict()
